/*=============================================================================
    Copyright (c) 2007-2008 Dan Marsden

    Distributed under the Boost Software License, Version 1.0. (See accompanying 
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/

/*=============================================================================
  Demo on simplified xml data structure.
  The data structure is recursive, and is built up from user defined structures,
  standard library data types and Boost.Variant.
  
  Demonstrate various manipulations on the xml tree, all of which are performed
  without writing any recursive traversal code. In fact they are not dependent
  on the exact data structure being used. Type information is used to identify
  which components of the tree need modification.

  The demonstration also shows how lambda functions can be constrained by the
  restrict<type> manipulator to only operate on certain types in the tree being
  traversed.
==============================================================================*/


#include <boost/variant.hpp>
#include <boost/fusion/adapted/struct/adapt_struct.hpp>

#include <boost/traversal/recursive/bu_full.hpp>
#include <boost/traversal/recursive/bu_full_parent.hpp>
#include <boost/traversal/functional/extend.hpp>
#include <boost/traversal/functional/restrict.hpp>

#include <boost/algorithm/string/case_conv.hpp>

#include <vector>
#include <string>
#include <iostream>

#include <boost/lambda/lambda.hpp>

namespace demo
{
    struct mini_xml;

    typedef boost::variant<
        boost::recursive_wrapper<mini_xml>,
        std::string> mini_xml_node;

    struct mini_xml
    {
        std::string name;
        std::vector<mini_xml_node> children;
        mini_xml(std::string const& name, std::vector<mini_xml_node> const& children = std::vector<mini_xml_node>())
            :
            name(name), children(children)
        {}
    };

    std::ostream& operator<<(std::ostream& stream, mini_xml const& mxml)
    {
        return stream << mxml.name << ' ' << mxml.children.size() << " children";
    }

    bool uppercase_strings(std::string& str)
    {
        boost::algorithm::to_upper(str);
        return false;
    }

    struct remove_private_data
    {
        template<typename Child, typename Parent>
        bool operator()(Child& child, Parent& parent) const
            {
                return false;
            }

        template<typename Parent>
        bool operator()(mini_xml& mxml, mini_xml_node& parent) const
            {
                if(mxml.name == "PRIVATE DATA")
                    parent = std::string("");
                return false;
            }
    };

    bool is_blank(mini_xml_node& node)
    {
        std::string* ptr = boost::get<std::string>(&node);
        return ptr && ptr->empty();
    }

    struct remove_blanks
    {
        template<typename T>
        bool operator()(T&) const
            {
                return false;
            }

        bool operator()(mini_xml& mxml) const
            {
                mxml.children.erase(
                    std::remove_if(
                        mxml.children.begin(), mxml.children.end(), is_blank), mxml.children.end());
                return false;
            }
    };
}

BOOST_FUSION_ADAPT_STRUCT(demo::mini_xml,
                          (std::string, name)
                          (std::vector<demo::mini_xml_node>, children))

int main()
{
    using namespace boost;
    using namespace boost::lambda;

    // Create a small demo xml data structure to experiment with
    demo::mini_xml mxml("demo_data");
    mxml.children.push_back(
        demo::mini_xml_node(
            demo::mini_xml(
                "public data", 
                std::vector<demo::mini_xml_node>(
                    1, 
                    demo::mini_xml("private data")))));

    mxml.children.push_back(
        demo::mini_xml_node(demo::mini_xml("private data")));

    // Enumerate all the xml nodes, bottom up, left to right, used restrict<type> to pick out only mini_xml 
    traversal::bu_full(
        mxml, traversal::restrict<bool(demo::mini_xml)>(std::cout << _1 << '\n'));

    // Convert all strings to upper case
    traversal::bu_full(
        mxml, traversal::extend(demo::uppercase_strings));

    // Show the modifications to the section headings (the text has been changed as well)
    traversal::bu_full(
        mxml, traversal::restrict<bool(demo::mini_xml)>(std::cout << _1 << '\n'));

    // Replace private data sections with blank strings
    traversal::bu_full_parent(
        mxml, demo::remove_private_data());

    // Remove all blank strings from the xml nodes
    traversal::bu_full(
        mxml, demo::remove_blanks());
}
