/*=============================================================================
    Copyright (c) 2007-2008 Dan Marsden

    Distributed under the Boost Software License, Version 1.0. (See accompanying 
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/
#include <boost/detail/lightweight_test.hpp>
#include <boost/traversal/functional/extend.hpp>
#include <boost/traversal/functional/false.hpp>

#include <string>
#include <vector>

using namespace boost;

namespace
{
    int increment(int& i)
    {
        return i + 1;
    }

    int sum2(int& i0, int& i1)
    {
        return i0 + i1;
    }

    int sum3(int& i0, int& i1, int& i2)
    {
        return i0 + i1 + i2;
    }

    void incr1(int& i0)
    {
        ++i0;
    }

    void incr2(int& i0, int& i1)
    {
        ++i0;
        ++i1;
    }

    void incr3(int& i0, int& i1, int& i2)
    {
        ++i0;
        ++i1;
        ++i2;
    }

    struct raise
    {
        typedef int result_type;
        typedef int& argument_type;

        explicit raise(int delta)
            : delta_(delta)
        {}

        int operator()(int& i0) const
        {
            return i0 + delta_;
        }

        int delta_;
    };

    struct sum_and_raise2
    {
        typedef int result_type;
        typedef int& first_argument_type;
        typedef int& second_argument_type;

        explicit sum_and_raise2(int delta)
            : delta_(delta)
        {}

        int operator()(int& i0, int& i1) const
        {
            return i0 + i1 + delta_;
        }

        int delta_;
    };

    struct sum_and_raise3
    {
        typedef int result_type;
        typedef int& first_argument_type;
        typedef int& second_argument_type;
        typedef int& third_argument_type;

        explicit sum_and_raise3(int delta)
            : delta_(delta)
        {}

        int operator()(int& i0, int& i1, int& i2) const
        {
            return i0 + i1 + i2 + delta_;
        }

        int delta_;
    };

    template<typename Func>
    void test_unary(Func f, int dflt = 0)
    {
        int arg0(101);
        BOOST_TEST(f(arg0) == arg0 + 1);

        std::string const badArg0("arg0");
        BOOST_TEST(f(badArg0) == dflt);
    }

    template<typename Func>
    void test_binary(Func f, int dflt = 0)
    {
        int arg0(101);
        int arg1(202);

        BOOST_TEST(f(arg0, arg1) == arg0 + arg1);

        std::string const badArg0("arg0");
        std::string const badArg1("arg1");
        BOOST_TEST(f(badArg0, badArg1) == dflt);
    }

    template<typename Func>
    void test_stateful_binary(Func f, int dflt = 0)
    {
        int arg0(101);
        int arg1(202);

        BOOST_TEST(f(arg0, arg1) == arg0 + arg1 + 1);

        std::string const badArg0("arg0");
        std::string const badArg1("arg1");
        BOOST_TEST(f(badArg0, badArg1) == dflt);
    }

    template<typename Func>
    void test_ternary(Func f, int dflt = 0)
    {
        int arg0(101);
        int arg1(202);
        int arg2(303);

        BOOST_TEST(f(arg0, arg1, arg2) == arg0 + arg1 + arg2);

        std::string const badArg0("arg0");
        std::string const badArg1("arg1");
        std::string const badArg2("arg2");
        BOOST_TEST(f(badArg0, badArg1, badArg2) == dflt);
    }

    template<typename Func>
    void test_stateful_ternary(Func f, int dflt = 0)
    {
        int arg0(101);
        int arg1(202);
        int arg2(303);

        BOOST_TEST(f(arg0, arg1, arg2) == arg0 + arg1 + arg2 + 1);

        std::string const badArg0("arg0");
        std::string const badArg1("arg1");
        std::string const badArg2("arg2");
        BOOST_TEST(f(badArg0, badArg1, badArg2) == dflt);
    }

    template<typename Func>
    void test_unary_void_return(Func f)
    {
        int i0 = 0;
        f(i0);
        BOOST_TEST(i0 == 1);

        std::string arg0("arg0");
        f(arg0);
        BOOST_TEST(arg0 == "arg0");
    }

    template<typename Func>
    void test_binary_void_return(Func f)
    {
        int i0 = 0, i1 = 0;
        f(i0, i1);
        BOOST_TEST(i0 == 1);
        BOOST_TEST(i1 == 1);

        std::string arg0("arg0");
        std::string arg1("arg1");
        f(arg0, arg1);
        BOOST_TEST(arg0 == "arg0");
        BOOST_TEST(arg1 == "arg1");
    }

    template<typename Func>
    void test_ternary_void_return(Func f)
    {
        int i0 = 0, i1 = 0, i2 = 0;
        f(i0, i1, i2);
        BOOST_TEST(i0 == 1);
        BOOST_TEST(i1 == 1);
        BOOST_TEST(i2 == 1);

        std::string arg0("arg0");
        std::string arg1("arg1");
        std::string arg2("arg2");
        f(arg0, arg1, arg2);
        BOOST_TEST(arg0 == "arg0");
        BOOST_TEST(arg1 == "arg1");
        BOOST_TEST(arg2 == "arg2");
    }
}

int main()
{
    test_unary(traversal::extend(increment));
    test_binary(traversal::extend(sum2));
    test_ternary(traversal::extend(sum3));

    int explicitDefault = 1;
    test_unary(traversal::extend(increment, explicitDefault), explicitDefault);
    test_binary(traversal::extend(sum2, explicitDefault), explicitDefault);
    test_ternary(traversal::extend(sum3, explicitDefault), explicitDefault);

    test_unary_void_return(traversal::extend(incr1));
    test_binary_void_return(traversal::extend(incr2));
    test_ternary_void_return(traversal::extend(incr3));

    test_unary(traversal::extend(raise(1)));
    test_stateful_binary(traversal::extend(sum_and_raise2(1)));
    test_stateful_ternary(traversal::extend(sum_and_raise3(1)));

    return boost::report_errors();
}
